__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Iterable, Optional

from . import BaseExecutableDriver, FlatRecursiveMixin
from ..enums import EmbeddingClsType

if False:
    from ..types.sets import DocumentSet


class BaseIndexDriver(FlatRecursiveMixin, BaseExecutableDriver):
    """Drivers inherited from this Driver will bind :meth:`add` by default """

    def __init__(
        self, executor: Optional[str] = None, method: str = 'add', *args, **kwargs
    ):
        super().__init__(executor, method, *args, **kwargs)

    def check_key_length(self, val: Iterable[str]):
        """
        Check if the max length of val(e.g. doc id) is larger than key_length.

        :param val: The values to be checked
        """
        m_val = max(len(v) for v in val)
        if m_val > self.exec.key_length:
            raise ValueError(
                f'{self.exec} allows only keys of length {self.exec.key_length}, '
                f'but yours is {m_val}.'
            )


class VectorIndexDriver(BaseIndexDriver):
    """Extracts embeddings and ids from the documents and forwards them to the executor.
    In case `method` is 'delete', the embeddings are ignored.
    If `method` is not 'delete', documents without content are filtered out.
    """

    @property
    def exec_embedding_cls_type(self) -> EmbeddingClsType:
        """Get the sparse class type of the attached executor.

        :return: Embedding class type of the attached executor, default value is `dense`
        """
        return EmbeddingClsType.from_string(self.exec.embedding_cls_type)

    def _get_documents_embeddings(self, docs: 'DocumentSet'):
        embedding_cls_type = self.exec_embedding_cls_type
        if embedding_cls_type.is_dense:
            return docs.all_embeddings
        else:
            return docs.get_all_sparse_embeddings(embedding_cls_type=embedding_cls_type)

    def _apply_all(self, docs: 'DocumentSet', *args, **kwargs) -> None:
        embed_vecs, docs_pts = self._get_documents_embeddings(docs)
        if docs_pts:
            keys = [doc.id for doc in docs_pts]
            self.check_key_length(keys)
            self.exec_fn(keys, embed_vecs)


class KVIndexDriver(BaseIndexDriver):
    """Forwards pairs of serialized documents and ids to the executor."""

    def _apply_all(self, docs: 'DocumentSet', *args, **kwargs) -> None:
        info = [(doc.id, doc.SerializeToString()) for doc in docs]
        if info:
            keys, values = zip(*info)
            self.check_key_length(keys)
            self.exec_fn(keys, values)


class DBMSIndexDriver(BaseIndexDriver):
    """Forwards ids, vectors, serialized Document to a BaseDBMSIndexer"""

    def _apply_all(self, docs: 'DocumentSet', *args, **kwargs) -> None:
        info = [
            (
                doc.id,
                doc.embedding,
                self._doc_without_embedding(doc).SerializeToString(),
            )
            for doc in docs
        ]
        if info:
            ids, vecs, metas = zip(*info)
            self.check_key_length(ids)
            self.exec_fn(ids, vecs, metas)

    @staticmethod
    def _doc_without_embedding(d):
        from .. import Document

        new_doc = Document(d, copy=True)
        new_doc.ClearField('embedding')
        return new_doc



from typing import Union, Optional, TypeVar, Dict

from google.protobuf import json_format

from ..mixin import ProtoTypeMixin
from ..sets import QueryLangSet
from ...enums import CompressAlgo, RequestType
from ...excepts import BadRequestType
from ...helper import random_identity, typename
from ...proto import jina_pb2

_body_type = set(str(v).lower() for v in RequestType)
_trigger_body_fields = set(
    kk
    for v in [
        jina_pb2.RequestProto.IndexRequestProto,
        jina_pb2.RequestProto.SearchRequestProto,
        jina_pb2.RequestProto.TrainRequestProto,
        jina_pb2.RequestProto.ControlRequestProto,
    ]
    for kk in v.DESCRIPTOR.fields_by_name.keys()
)
_trigger_req_fields = set(
    jina_pb2.RequestProto.DESCRIPTOR.fields_by_name.keys()
).difference(_body_type)
_trigger_fields = _trigger_req_fields.union(_trigger_body_fields)

__all__ = ['Request', 'Response']

RequestSourceType = TypeVar(
    'RequestSourceType', jina_pb2.RequestProto, bytes, str, Dict
)


class Request(ProtoTypeMixin):
    """
    :class:`Request` is one of the **primitive data type** in Jina.

    It offers a Pythonic interface to allow users access and manipulate
    :class:`jina.jina_pb2.RequestProto` object without working with Protobuf itself.

    A container for serialized :class:`jina_pb2.RequestProto` that only triggers deserialization
    and decompression when receives the first read access to its member.

    It overrides :meth:`__getattr__` to provide the same get/set interface as an
    :class:`jina_pb2.RequestProto` object.

    :param request: The request.
    :param envelope: EnvelopeProto object.
    :param copy: Copy the request if ``copy`` is True.
    """

    def __init__(
        self,
        request: Optional[Union[bytes, dict, str, 'jina_pb2.RequestProto']] = None,
        envelope: Optional['jina_pb2.EnvelopeProto'] = None,
        copy: bool = False,
    ):
        """
        Set constructor method.

        :param request: request object as bytes, dictionary, string or protobuf instance
        :param envelope: envelope of the request
        :param copy: if true, request is copied
        """
        self._buffer = None
        self._pb_body = jina_pb2.RequestProto()  # type: 'jina_pb2.RequestProto'
        try:
            if isinstance(request, jina_pb2.RequestProto):
                if copy:
                    self._pb_body.CopyFrom(request)
                else:
                    self._pb_body = request
            elif isinstance(request, dict):
                json_format.ParseDict(request, self._pb_body)
            elif isinstance(request, str):
                json_format.Parse(request, self._pb_body)
            elif isinstance(request, bytes):
                self._buffer = request
                self._pb_body = None
            elif request is None:
                # make sure every new request has a request id
                self._pb_body.request_id = random_identity()
            elif request is not None:
                # note ``None`` is not considered as a bad type
                raise ValueError(f'{typename(request)} is not recognizable')
        except Exception as ex:
            raise BadRequestType(f'fail to construct a request from {request}') from ex

        self._envelope = envelope
        self.is_used = False  #: Return True when request has been r/w at least once

    def __getattr__(self, name: str):
        # https://docs.python.org/3/reference/datamodel.html#object.__getattr__
        if name in _trigger_body_fields:
            return getattr(self.body, name)
        else:
            return getattr(self.proto, name)

    @property
    def body(self):
        """
        Return the request type, raise ``ValueError`` if request_type not set.

        :return: body property
        """
        if self._request_type:
            return getattr(self.proto, self._request_type)
        else:
            raise ValueError(f'"request_type" is not set yet')

    @property
    def _request_type(self) -> str:
        return self.proto.WhichOneof('body')

    @property
    def request_type(self) -> Optional[str]:
        """
        Return the request body type, when not set yet, return ``None``.

        :return: request type
        """
        if self._request_type:
            return self.body.__class__.__name__

    def as_typed_request(self, request_type: str):
        """
        Change the request class according to the one_of value in ``body``.

        :param request_type: string representation of the request type
        :return: self
        """
        from .train import TrainRequest
        from .search import SearchRequest
        from .control import ControlRequest
        from .index import IndexRequest
        from .delete import DeleteRequest
        from .update import UpdateRequest
        from .dump import DumpRequest

        rt = request_type.upper()
        if rt.startswith(str(RequestType.TRAIN)):
            self.__class__ = TrainRequest
        elif rt.startswith(str(RequestType.DELETE)):
            self.__class__ = DeleteRequest
        elif rt.startswith(str(RequestType.INDEX)):
            self.__class__ = IndexRequest
        elif rt.startswith(str(RequestType.SEARCH)):
            self.__class__ = SearchRequest
        elif rt.startswith(str(RequestType.UPDATE)):
            self.__class__ = UpdateRequest
        elif rt.startswith(str(RequestType.CONTROL)):
            self.__class__ = ControlRequest
        elif rt.startswith(str(RequestType.DUMP)):
            self.__class__ = DumpRequest
        else:
            raise TypeError(f'{request_type} is not recognized')
        return self

    @request_type.setter
    def request_type(self, value: str):
        """
        Set the type of this request, but keep the body empty.

        :param value: string representation of request type
        """
        value = value.lower()
        if value in _body_type:
            getattr(self.proto, value).SetInParent()
        else:
            raise ValueError(f'{value} is not valid, must be one of {_body_type}')
        self.as_typed_request(self._request_type)

    @staticmethod
    def _decompress(data: bytes, algorithm: str) -> bytes:
        if not algorithm:
            return data

        ctag = CompressAlgo.from_string(algorithm)
        if ctag == CompressAlgo.LZ4:
            import lz4.frame

            data = lz4.frame.decompress(data)
        elif ctag == CompressAlgo.BZ2:
            import bz2

            data = bz2.decompress(data)
        elif ctag == CompressAlgo.LZMA:
            import lzma

            data = lzma.decompress(data)
        elif ctag == CompressAlgo.ZLIB:
            import zlib

            data = zlib.decompress(data)
        elif ctag == CompressAlgo.GZIP:
            import gzip

            data = gzip.decompress(data)
        return data

    @property
    def proto(self) -> 'jina_pb2.RequestProto':
        """
        Cast ``self`` to a :class:`jina_pb2.RequestProto`. This will trigger
        :attr:`is_used`. Laziness will be broken and serialization will be recomputed when calling
        :meth:`SerializeToString`.

        :return: protobuf instance
        """
        if self._pb_body:
            # if request is already given while init
            self.is_used = True
            return self._pb_body
        else:
            # if not then build one from buffer
            r = jina_pb2.RequestProto()
            _buffer = self._decompress(
                self._buffer,
                self._envelope.compression.algorithm if self._envelope else None,
            )
            r.ParseFromString(_buffer)
            self.is_used = True
            self._pb_body = r
            # # Though I can modify back the envelope, not sure if it is a good design:
            # # My intuition is: if the content is changed dramatically, e.g. from index to control request,
            # # then whatever writes on the envelope should be dropped
            # # depreciated. The only reason to reuse the envelope is saving cost on Envelope(), which is
            # # really a minor minor (and evil) optimization.
            # if self._envelope:
            #     self._envelope.request_type = getattr(r, r.WhichOneof('body')).__class__.__name__
            return r

    def SerializeToString(self) -> bytes:
        """
        Convert serialized data to string.

        :return: serialized request
        """
        if self.is_used:
            return self.proto.SerializeToString()
        else:
            # no touch, skip serialization, return original
            return self._buffer

    @property
    def queryset(self) -> 'QueryLangSet':
        """
        Get the queryset in :class:`QueryLangSet` type.

        :return: query lang set
        """
        self.is_used = True
        return QueryLangSet(self.proto.queryset)

    def as_response(self):
        """
        Return a weak reference of this object but as :class:`Response` object. It gives a more
        consistent semantics on the client.
        """
        base_cls = self.__class__
        base_cls_name = self.__class__.__name__
        self.__class__ = type(base_cls_name, (base_cls, Response), {})


class Response:
    """
    Response is the :class:`Request` object returns from the flow. Right now it shares the same representation as
    :class:`Request`. At 0.8.12, :class:`Response` is a simple alias. But it does give a more consistent semantic on
    the client API: send a :class:`Request` and receive a :class:`Response`.
    """



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Dict, Optional, List

import numpy as np

from .. import BaseExecutor


class BaseRanker(BaseExecutor):
    """
    The base class for a `Ranker`
    :param query_required_keys: Set of keys or features to be extracted from query `Document` by the `Driver` so that
        they are passed as query features or metainfo.
    :param match_required_keys: Set of keys or features to be extracted from match `Document` by the `Driver` so that
        they are passed as match features or metainfo.
    :param args: Extra positional arguments
    :param kwargs: Extra keyword arguments

    .. note::
        See how the attributes are accessed in :class:`Document` in :meth:`get_attrs`.

        .. highlight:: python
        .. code-block:: python

            query = Document({'tags': {'color': 'blue'})
            match = Document({'tags': {'color': 'blue', 'price': 1000}})

            ranker = BaseRanker(query_required_keys=('tags__color'), match_required_keys=('tags__color, 'tags__price')
    """

    def __init__(
        self,
        query_required_keys: Optional[List[str]] = None,
        match_required_keys: Optional[List[str]] = None,
        *args,
        **kwargs
    ):
        """

        :param query_required_keys: Set of keys or features to be extracted from query `Document` by the `Driver` so that
            they are passed as query features or metainfo.
        :param match_required_keys: Set of keys or features to be extracted from match `Document` by the `Driver` so that
            they are passed as match features or metainfo.
        :param args: Extra positional arguments
        :param kwargs: Extra keyword arguments

        .. note::
            See how the attributes are accessed in :class:`Document` in :meth:`get_attrs`.

            .. highlight:: python
            .. code-block:: python

                query = Document({'tags': {'color': 'blue'})
                match = Document({'tags': {'color': 'blue', 'price': 1000}})

                ranker = BaseRanker(query_required_keys=('tags__color'), match_required_keys=('tags__color, 'tags__price')
        """
        super().__init__(*args, **kwargs)
        self.query_required_keys = query_required_keys
        self.match_required_keys = match_required_keys

    def score(self, *args, **kwargs):
        """Calculate the score. Base class method needs to be implemented in subclass.
        :param args: Extra positional arguments
        :param kwargs: Extra keyword arguments
        """
        raise NotImplementedError


class Chunk2DocRanker(BaseRanker):
    """A :class:`Chunk2DocRanker` translates the chunk-wise score (distance) to the doc-wise score.

    In the query-time, :class:`Chunk2DocRanker` is an almost-always required component.
    Because in the end we want to retrieve top-k documents of given query-document not top-k chunks of
    given query-chunks. The purpose of :class:`Chunk2DocRanker` is to aggregate the already existed top-k chunks
    into documents.

    The key function here is :func:`score`.

    .. seealso::
        :mod:`jina.drivers.handlers.score`

    """

    COL_PARENT_ID = 'match_parent_id'
    COL_DOC_CHUNK_ID = 'match_doc_chunk_id'
    COL_QUERY_CHUNK_ID = 'match_query_chunk_id'
    COL_SCORE = 'score'

    def score(
        self, match_idx: 'np.ndarray', query_chunk_meta: Dict, match_chunk_meta: Dict
    ) -> float:
        """
        Given a set of queries (that may correspond to the chunks of a root level query) and a set of matches
        corresponding to the same parent id, compute the matching score of the common parent of the set of matches.
        Returns a score corresponding to the score of the parent document of the matches in `match_idx`

        :param match_idx: A [N x 4] numpy ``ndarray``, column-wise:
                - ``match_idx[:, 0]``: ``parent_id`` of the matched docs, integer
                - ``match_idx[:, 1]``: ``id`` of the matched chunks, integer
                - ``match_idx[:, 2]``: ``id`` of the query chunks, integer
                - ``match_idx[:, 3]``: distance/metric/score between the query and matched chunks, float.
                All the matches belong to the same `parent`
        :param query_chunk_meta: The meta information of the query chunks, where the key is query chunks' ``chunk_id``,
            the value is extracted by the ``query_required_keys``.
        :param match_chunk_meta: The meta information of the matched chunks, where the key is matched chunks'
            ``chunk_id``, the value is extracted by the ``match_required_keys``.


        TODO:
        - ``match_idx[:, 0]`` is redundant because all the matches have the same ``parent_id``.

        """
        raise NotImplementedError


class Match2DocRanker(BaseRanker):
    """
    Re-scores the matches for a document. This Ranker is only responsible for
    calculating new scores and not for the actual sorting. The sorting is handled
    in the respective ``Matches2DocRankDriver``.

    Possible implementations:
        - ReverseRanker (reverse scores of all matches)
        - BucketShuffleRanker (first buckets matches and then sort each bucket).
    """

    COL_MATCH_ID = 'match_doc_chunk_id'
    COL_SCORE = 'score'

    def score(
        self,
        old_matches_scores: List[List[float]],
        queries_metas: List[Dict],
        matches_metas: List[List[Dict]],
    ) -> List[List[float]]:
        """
        Calculates the new scores for matches and returns them. Returns an iterable of the scores to be assigned to the matches.
        The returned scores need to be returned in the same order as the input `:param old_match_scores`.

        .. note::
            The length of `old_match_scores`, `queries_metas` and `matches_metas` correspond to the amount of queries in the batch for which
            one wants to score its matches.

            Every Sequence in match metas correspond to the amount of retrieved matches per query.

            The resulting list of scores will provide a list of score for every query. And every list will be ordered in the same way as the `matches_metas` lists

        :param old_matches_scores: Contains old scores in a list for every query
        :param queries_metas: List of dictionaries containing all the query meta information requested by the `query_required_keys` class_variable for each query in a batch.
        :param matches_metas: List of lists containing all the matches meta information requested by the `match_required_keys` class_variable for every query. Sorted in the same way as `old_match_scores`
        """
        raise NotImplementedError



from ..sets import DocumentSet
from ...proto import jina_pb2


class DocsPropertyMixin:
    """Mixin class of docs property."""

    @property
    def docs(self) -> 'DocumentSet':
        """Get the :class: `DocumentSet` with sequence `body.docs` as content."""
        self.is_used = True
        return DocumentSet(self.body.docs)


class GroundtruthPropertyMixin:
    """Mixin class of groundtruths property."""

    @property
    def groundtruths(self) -> 'DocumentSet':
        """Get the groundtruths in :class: `DocumentSet` type."""
        self.is_used = True
        return DocumentSet(self.body.groundtruths)


class IdsMixin:
    """Mixin class of ids property."""

    @property
    def ids(self):
        """Get the ids."""
        return self.body.ids


class CommandMixin:
    """Mixin class of command property."""

    @property
    def command(self) -> str:
        """Get the command."""
        self.is_used = True
        return jina_pb2.RequestProto.ControlRequestProto.Command.Name(
            self.proto.control.command
        )



"""Argparser module for hub new"""
from ..helper import add_arg_group


def mixin_hub_new_parser(parser):
    """Add the options for `hub new` to the parser
    :param parser: the parser
    """
    gp = add_arg_group(parser, title='Create')

    gp.add_argument(
        '--output-dir',
        type=str,
        default='.',
        help='where to output the generated project dir into.',
    )
    gp.add_argument(
        '--template',
        type=str,
        default='https://github.com/jina-ai/cookiecutter-jina-hub.git',
        help='The cookiecutter template directory containing a project template directory, or a URL to a git repository. Only used when `--type template`',
    )
    gp.add_argument(
        '--type',
        type=str,
        default='pod',
        choices=['pod', 'app', 'template'],
        help='The template type for executor hub pod or app using cookiecutter.',
    )
    gp.add_argument(
        '--overwrite',
        action='store_true',
        default=False,
        help='overwrite the contents of output directory if it exists',
    )



from typing import Union, Sequence

from ...clients.base import CallbackFnType


class AsyncControlFlowMixin:
    """The asynchronous version of the Mixin for controlling, scaling the Flow"""

    async def reload(
        self,
        targets: Union[str, Sequence[str]],
        on_done: CallbackFnType = None,
        on_error: CallbackFnType = None,
        on_always: CallbackFnType = None,
        **kwargs,
    ):
        """Reload the executor of certain peas/pods in the Flow
        It will start a :py:class:`CLIClient` and call :py:func:`reload`.

        :param targets: the regex string or list of regex strings to match the pea/pod names.
        :param on_done: the function to be called when the :class:`Request` object is resolved.
        :param on_error: the function to be called when the :class:`Request` object is rejected.
        :param on_always: the function to be called when the :class:`Request` object is  is either resolved or rejected.
        :param kwargs: accepts all keyword arguments of `jina client` CLI
        :yield: result
        """
        async for r in self._get_client(**kwargs).reload(
            targets, on_done, on_error, on_always, **kwargs
        ):
            yield r



__copyright__ = "Copyright (c) 2021 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import os
from typing import Optional

import numpy as np

from jina.executors.decorators import batching, as_ndarray
from jina.executors.encoders.frameworks import BaseTorchEncoder
from jina.excepts import PretrainedModelFileDoesNotExist


class CustomImageTorchEncoder(BaseTorchEncoder):
    """
    :class:`CustomImageTorchEncoder` encodes ``Document`` content from a ndarray,
    potentially B x (Channel x Height x Width) into a ndarray of `B x D`.

    Internally, :class:`CustomImageTorchEncoder` wraps any custom torch
    model not part of models from `torchvision.models`.
    https://pytorch.org/docs/stable/torchvision/models.html

    :param model_path: The path where the model is stored.
    :param layer_name: Name of the layer from where to extract the feature map.
    :param pool_strategy: Pooling strategy for the encoder. Default is `mean`.
    :param channel_axis: The axis of the channel, default is 1.
    :param args:  Additional positional arguments.
    :param kwargs: Additional keyword arguments.
    """

    def __init__(self, model_path: Optional[str] = 'models/mobilenet_v2.pth',
                 layer_name: Optional[str] = 'features',
                 pool_strategy: str = 'mean',
                 channel_axis: int = 1,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model_path = model_path
        self.layer_name = layer_name
        self.channel_axis = channel_axis
        # axis 0 is the batch
        self._default_channel_axis = 1
        self.pool_strategy = pool_strategy

    def post_init(self):
        """Load model. Will raise an exception if the model doesn't exist."""
        super().post_init()
        if self.model_path and os.path.exists(self.model_path):
            import torch
            if self.pool_strategy is not None:
                self.pool_fn = getattr(np, self.pool_strategy)
            self.model = torch.load(self.model_path)
            self.model.eval()
            self.to_device(self.model)
            self.layer = getattr(self.model, self.layer_name)
        else:
            raise PretrainedModelFileDoesNotExist(f'model {self.model_path} does not exist')

    def _get_features(self, content):
        feature_map = None

        def get_activation(model, model_input, output):
            nonlocal feature_map
            feature_map = output.detach()

        handle = self.layer.register_forward_hook(get_activation)
        self.model(content)
        handle.remove()
        return feature_map

    def _get_pooling(self, feature_map: 'np.ndarray') -> 'np.ndarray':
        if feature_map.ndim == 2 or self.pool_strategy is None:
            return feature_map
        return self.pool_fn(feature_map, axis=(2, 3))

    @batching
    @as_ndarray
    def encode(self, content: 'np.ndarray', *args, **kwargs) -> 'np.ndarray':
        """
        Encode input data into `np.ndarray` of size `B x D`.
        Where `B` is the batch size and `D` is the Dimension.

        :param content: An array in size `B`.
        :param args:  Additional positional arguments.
        :param kwargs: Additional keyword arguments.
        :return: An ndarray in size `B x D`.
        """
        if self.channel_axis != self._default_channel_axis:
            content = np.moveaxis(content, self.channel_axis, self._default_channel_axis)
        import torch
        _input = torch.from_numpy(content.astype('float32'))
        if self.on_gpu:
            _input = _input.cuda()
        _feature = self._get_features(content=_input).detach()
        if self.on_gpu:
            _feature = _feature.cpu()
        _feature = _feature.numpy()
        return self._get_pooling(_feature)



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Optional, Any

from . import BaseExecutableDriver, FlatRecursiveMixin, DocsExtractUpdateMixin

# noinspection PyUnreachableCode
if False:
    from .. import Document
    from .. import DocumentSet


class BaseEncodeDriver(BaseExecutableDriver):
    """Drivers inherited from this Driver will bind :meth:`encode` by default """

    def __init__(
        self, executor: Optional[str] = None, method: str = 'encode', *args, **kwargs
    ):
        super().__init__(executor, method, *args, **kwargs)


class EncodeDriver(DocsExtractUpdateMixin, FlatRecursiveMixin, BaseEncodeDriver):
    """Extract the content from documents and call executor and do encoding"""

    def update_single_doc(self, doc: 'Document', exec_result) -> None:
        """Update the document embedding with returned ndarray result

        :param doc: the Document object
        :param exec_result: the single result from :meth:`exec_fn`
        """
        doc.embedding = exec_result


class ScipySparseEncodeDriver(
    DocsExtractUpdateMixin, FlatRecursiveMixin, BaseEncodeDriver
):
    """Extract the content from documents and call executor and do encoding"""

    def update_docs(self, docs_pts: 'DocumentSet', exec_results: Any) -> None:
        """Update the document embedding with returned sparse matrix

        :param: docs_pts: the set of document to be updated
        :param: exec_results: the results from :meth:`exec_fn`
        """
        for idx, doc in enumerate(docs_pts):
            doc.embedding = exec_results.getrow(idx)



from typing import Generator

import numpy as np

from jina.executors.indexers.dump import import_vectors
from jina.executors.indexers.query import BaseQueryIndexer
from jina.executors.indexers.vector import NumpyIndexer


class NumpyQueryIndexer(NumpyIndexer, BaseQueryIndexer):
    """An exhaustive vector indexers implemented with numpy and scipy.

    .. note::
        Metrics other than `cosine` and `euclidean` requires ``scipy`` installed.

    :param metric: The distance metric to use. `braycurtis`, `canberra`, `chebyshev`, `cityblock`, `correlation`,
                    `cosine`, `dice`, `euclidean`, `hamming`, `jaccard`, `jensenshannon`, `kulsinski`,
                    `mahalanobis`,
                    `matching`, `minkowski`, `rogerstanimoto`, `russellrao`, `seuclidean`, `sokalmichener`,
                    `sokalsneath`, `sqeuclidean`, `wminkowski`, `yule`.
    :param backend: `numpy` or `scipy`, `numpy` only supports `euclidean` and `cosine` distance
    :param compress_level: compression level to use
    """

    def _load_dump(self, dump_path):
        """Load the dump at the path

        :param dump_path: the path of the dump"""
        ids, vecs = import_vectors(dump_path, str(self.pea_id))
        self._add(ids, vecs)
        self.write_handler.flush()
        self.write_handler.close()
        self.handler_mutex = False
        self.is_handler_loaded = False
        test_vecs = np.array([np.random.random(self.num_dim)], dtype=self.dtype)
        assert self.query(test_vecs, 1) is not None

    def _add(self, keys: Generator, vectors: Generator, *args, **kwargs) -> None:
        """Add the embeddings and document ids to the index.

        NOTE::

            This replaces the parent class' `_add` since we
            need to adapt to use Generators from the dump loading

        :param keys: a list of ``id``, i.e. ``doc.id`` in protobuf
        :param vectors: embeddings
        :param args: not used
        :param kwargs: not used
        """
        keys = np.array(list(keys), (np.str_, self.key_length))
        vectors_nr = 0
        for vector in vectors:
            if not getattr(self, 'num_dim', None):
                self.num_dim = vector.shape[0]
                self.dtype = vector.dtype.name
            self.write_handler.write(vector.tobytes())
            vectors_nr += 1

        if vectors_nr != keys.shape[0]:
            raise ValueError(
                f'Different number of vectors and keys. {vectors_nr} vectors and {len(keys)} keys. Validate your dump'
            )

        self.valid_indices = np.concatenate(
            (self.valid_indices, np.full(len(keys), True))
        )
        self.key_bytes += keys.tobytes()
        self._size += keys.shape[0]



import os

from jina import Segmenter, Crafter
from jina.executors.decorators import single


class SimpleCrafter(Crafter):
    """Simple crafter for multimodal example."""

    @single
    def craft(self, tags):
        """
        Read the data and add tags.

        :param tags: tags of data
        :return: crafted data
        """
        return {
            'text': tags['caption'],
            'uri': f'{os.environ["HW_WORKDIR"]}/people-img/{tags["image"]}',
        }


class BiSegmenter(Segmenter):
    """Segmenter for multimodal example."""

    @single(slice_nargs=2)
    def segment(self, text, uri):
        """
        Segment data into text and uri.

        :param text: text data
        :param uri: uri data of images
        :return: Segmented data.
        """
        return [
            {'text': text, 'mime_type': 'text/plain'},
            {'uri': uri, 'mime_type': 'image/jpeg'},
        ]



__copyright__ = "Copyright (c) 2021 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Tuple, Optional

import numpy as np
from jina.executors.devices import FaissDevice
from jina.executors.indexers.vector import BaseNumpyIndexer
from jina.executors.decorators import batching


class FaissIndexer(FaissDevice, BaseNumpyIndexer):
    batch_size = 512

    """Faiss powered vector indexer

    For more information about the Faiss supported parameters and installation problems, please consult:
        - https://github.com/facebookresearch/faiss

    .. note::
        Faiss package dependency is only required at the query time.
    """

    def __init__(self,
                 index_key: str,
                 train_filepath: Optional[str] = None,
                 max_num_training_points: Optional[int] = None,
                 requires_training: bool = True,
                 distance: str = 'l2',
                 normalize: bool = False,
                 nprobe: int = 1,
                 *args,
                 **kwargs):
        """
        Initialize an Faiss Indexer

        :param index_key: index type supported by ``faiss.index_factory``
        :param train_filepath: the training data file path, e.g ``faiss.tgz`` or `faiss.npy`. The data file is expected
            to be either `.npy` file from `numpy.save()` or a `.tgz` file from `NumpyIndexer`. If none is provided, `indexed` data will be used
            to train the Indexer (In that case, one must be careful when sharding is enabled, because every shard will be trained with its own part of data).
            The data will only be loaded if `requires_training` is set to True.
        :param max_num_training_points: Optional argument to consider only a subset of training points to training data from `train_filepath`.
            The points will be selected randomly from the available points
        :param requires_training: Boolean flag indicating if the index type requires training to be run before building index.
        :param distance: 'l2' or 'inner_product' accepted. Determines which distances to optimize by FAISS. l2...smaller is better, inner_product...larger is better
        :param normalize: whether or not to normalize the vectors e.g. for the cosine similarity https://github.com/facebookresearch/faiss/wiki/MetricType-and-distances#how-can-i-index-vectors-for-cosine-similarity
        :param nprobe: Number of clusters to consider at search time.

        .. highlight:: python
        .. code-block:: python
            # generate a training file in `.tgz`
            import gzip
            import numpy as np
            from jina.executors.indexers.vector.faiss import FaissIndexer

            train_filepath = 'faiss_train.tgz'
            train_data = np.random.rand(10000, 128)
            with gzip.open(train_filepath, 'wb', compresslevel=1) as f:
                f.write(train_data.astype('float32'))
            indexer = FaissIndexer('PCA64,FLAT', train_filepath)

            # generate a training file in `.npy`
            train_filepath = 'faiss_train'
            np.save(train_filepath, train_data)
            indexer = FaissIndexer('PCA64,FLAT', train_filepath)
        """
        super().__init__(*args, **kwargs)
        self.index_key = index_key
        self.requires_training = requires_training
        self.train_filepath = train_filepath if self.requires_training else None
        self.max_num_training_points = max_num_training_points
        self.distance = distance
        self.normalize = normalize
        self.nprobe = nprobe

    def build_advanced_index(self, vecs: 'np.ndarray'):
        """Build an advanced index structure from a numpy array.

        :param vecs: numpy array containing the vectors to index
        """
        import faiss

        metric = faiss.METRIC_L2
        if self.distance == 'inner_product':
            self.logger.warning('inner_product will be output as distance instead of similarity.')
            metric = faiss.METRIC_INNER_PRODUCT
        if self.distance not in {'inner_product', 'l2'}:
            self.logger.warning('Invalid distance metric for Faiss index construction. Defaulting to l2 distance')

        index = self.to_device(index=faiss.index_factory(self.num_dim, self.index_key, metric))

        if self.requires_training:
            if self.train_filepath:
                train_data = self._load_training_data(self.train_filepath)
            else:
                self.logger.info(f'Taking indexed data as training points')
                train_data = vecs
            if train_data is None:
                self.logger.warning('Loading training data failed. some faiss indexes require previous training.')
            else:
                if self.max_num_training_points:
                    self.logger.warning(f'From train_data with num_points {train_data.shape[0]}, '
                                        f'sample {self.max_num_training_points} points')
                    random_indices = np.random.choice(train_data.shape[0],
                                                      size=min(self.max_num_training_points, train_data.shape[0]),
                                                      replace=False)
                    train_data = train_data[random_indices, :]
                train_data = train_data.astype(np.float32)
                if self.normalize:
                    faiss.normalize_L2(train_data)
                self._train(index, train_data)

        self._build_partial_index(vecs, index)
        index.nprobe = self.nprobe
        return index

    @batching
    def _build_partial_index(self, vecs: 'np.ndarray', index):
        vecs = vecs.astype(np.float32)
        if self.normalize:
            from faiss import normalize_L2
            normalize_L2(vecs)
        index.add(vecs)

    def query(self, vecs: 'np.ndarray', top_k: int, *args, **kwargs) -> Tuple['np.ndarray', 'np.ndarray']:
        """Find the top-k vectors with smallest ``metric`` and return their ids in ascending order.
        :param keys: numpy array containing vectors to search for
        :param top_k: upper limit of responses for each search vector
        """
        if self.normalize:
            from faiss import normalize_L2
            normalize_L2(vecs)
        dist, ids = self.query_handler.search(vecs, top_k)
        if self.distance == 'inner_product':
            dist = 1 - dist
        keys = self._int2ext_id[self.valid_indices][ids]
        return keys, dist

    def _train(self, index, data: 'np.ndarray', *args, **kwargs) -> None:
        _num_samples, _num_dim = data.shape
        if not self.num_dim:
            self.num_dim = _num_dim
        if self.num_dim != _num_dim:
            raise ValueError('training data should have the same number of features as the index, {} != {}'.format(
                self.num_dim, _num_dim))
        self.logger.info(f'Training faiss Indexer with {_num_samples} points of {self.num_dim}')

        index.train(data)

    def _load_training_data(self, train_filepath: str) -> 'np.ndarray':
        self.logger.info(f'Loading training data from {train_filepath}')
        result = None
        try:
            result = self._load_gzip(train_filepath)
        except Exception as e:
            self.logger.error('Loading training data from gzip failed, filepath={}, {}'.format(train_filepath, e))

        if result is None:
            try:
                result = np.load(train_filepath)
                if isinstance(result, np.lib.npyio.NpzFile):
                    self.logger.warning('.npz format is not supported. Please save the array in .npy format.')
                    result = None
            except Exception as e:
                self.logger.error(
                    'Loading training data with np.load failed, filepath={}, {}'.format(train_filepath, e))

        if result is None:
            try:
                # Read from binary file:
                with open(train_filepath, 'rb') as f:
                    result = f.read()
            except Exception as e:
                self.logger.error(
                    'Loading training data from binary file failed, filepath={}, {}'.format(train_filepath, e))
        return result



__copyright__ = "Copyright (c) 2021 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import typing
import numpy as np
from scipy.sparse import coo_matrix, csr_matrix, bsr_matrix, csc_matrix
from pysparnn.matrix_distance import (
    CosineDistance,
    UnitCosineDistance,
    SlowEuclideanDistance,
    DenseCosineDistance,
)

from jina.executors.indexers.vector import BaseVectorIndexer

SparseMatrixType = typing.Union[csr_matrix, coo_matrix, bsr_matrix, csc_matrix]


def check_indexer(func):
    def checker(self, *args, **kwargs):
        if self.multi_cluster_index:
            raise ValueError('Not possible query while indexing')
        else:
            return func(self, *args, **kwargs)

    return checker


class PysparnnIndexer(BaseVectorIndexer):
    """
    :class:`PysparnnIndexer` Approximate Nearest Neighbor Search for Sparse Data in Python using PySparNN.

    For more information about the Pysparnn supported parameters and installation please consult:
        - https://github.com/facebookresearch/pysparnn

    """

    embedding_cls_type = 'scipy_csr'

    def __init__(
        self,
        k_clusters: int = 2,
        metric: str = 'cosine',
        num_indexes: int = 2,
        prefix_filename: str = 'pysparnn_index',
        *args,
        **kwargs,
    ):
        """Initializes a PysparnnIndexer Indexer

        :param k_clusters: number of clusters to be used in the multi_cluster_index from Pysparnn.
        :param metric: type of metric used for query, one from ['cosine', 'unit_cosine'. 'euclidean','dense_cosine']
        :param num_indexes: number of indexses used in the multi_cluster_index from Pysparnn.
        :param prefix_filename: prefix used when storing indices to disk
        :param args: additional parameters.
        :param kwargs: additional positional parameters.
        """
        super().__init__(*args, **kwargs)
        self.index = {}
        self.metric = self._assign_distance_class(metric)
        self.k_clusters = k_clusters
        self.num_indexes = num_indexes
        self.multi_cluster_index = None
        self.index_filename = prefix_filename + '_index.npz'
        self.indices_filename = prefix_filename + '_indices.npy'

    def get_add_handler(self):
        pass

    def get_create_handler(self):
        pass

    def get_write_handler(self):
        pass

    def get_query_handler(self):
        pass

    def post_init(self) -> None:
        """Load index if exist."""
        import os

        super().post_init()
        if os.path.exists(self.index_abspath):
            self._load_index_from_disk()

    def close(self) -> None:
        """
        Release the resources as executor is destroyed, need to be overridden
        """
        self._store_index_to_disk()
        super().close()

    def _assign_distance_class(self, metric: str):

        if metric == 'cosine':
            class_metric = CosineDistance
        elif metric == 'unit_cosine':
            class_metric = UnitCosineDistance
        elif metric == 'euclidean':
            class_metric = SlowEuclideanDistance
        elif metric == 'dense_cosine':
            class_metric = DenseCosineDistance
        else:
            raise ValueError(f'metric={metric} is not a valid metric')

        return class_metric

    def build_advanced_index(self) -> None:
        """Build the index using pysparnn `cluster_index` and stores it in `multi_cluster_index` """

        import pysparnn.cluster_index as ci
        import scipy

        if not self.index:
            raise ValueError(
                'Index is empty, please add data into the indexer using `add` method.'
            )
        keys = []
        indexed_vectors = []
        for key, vector in self.index.items():
            keys.append(key)
            indexed_vectors.append(vector)

        self.multi_cluster_index = ci.MultiClusterIndex(
            features=scipy.sparse.vstack(indexed_vectors),
            records_data=keys,
            distance_type=self.metric,
            num_indexes=self.num_indexes,
        )

    def query(self, vectors: SparseMatrixType, top_k: int, *args, **kwargs):
        """Find the top-k vectors with smallest ``metric`` and return their ids in ascending order.

        :return: a tuple of two ndarrays.
            The first array contains indices, the second array contains distances.
            If `n_vectors = vector.shape[0]` both arrays have shape `n_vectors x top_k`

        :param vectors: the vectors with which to search
        :param top_k: number of results to return
        :param args: additional positional parameters.
        :param kwargs: additional positional parameters.
        :return: tuple of arrays of the form `(indices, distances`
        """

        if not self.multi_cluster_index:
            self.build_advanced_index()

        index_distance_pairs = self.multi_cluster_index.search(
            vectors, k=top_k, k_clusters=self.k_clusters, return_distance=True
        )
        distances = []
        indices = []
        for record in index_distance_pairs:
            distance_list, index_list = zip(*record)
            distances.append(distance_list)
            indices.append(index_list)
        return np.array(indices), np.array(distances).astype(np.float)

    @check_indexer
    def add(
        self, keys: typing.List, vectors: SparseMatrixType, *args, **kwargs
    ) -> None:
        """Add keys and vectors to the indexer.

        :param keys: keys associated to the vectors
        :param vectors: vectors with which to search
        :param args: additional positional parameters.
        :param kwargs: additional positional parameters.
        """
        for key, vector in zip(keys, vectors):
            self.index[key] = vector

    @check_indexer
    def update(
        self, keys: typing.List, vectors: SparseMatrixType, *args, **kwargs
    ) -> None:
        """Update the embeddings on the index via document ids (keys).

        :param keys: keys associated to the vectors
        :param vectors: vectors with which to search
        :param args: additional positional parameters.
        :param kwargs: additional positional parameters.
        """

        for key, vector in zip(keys, vectors):
            self.index[key] = vector

    @check_indexer
    def delete(self, keys: typing.List, *args, **kwargs) -> None:
        """Delete the embeddings from the index via document ids (keys).

        :param keys: a list of ids
        :param args: additional positional parameters.
        :param kwargs: additional positional parameters.
        """
        for key in keys:
            self.index.pop(key)

    def _store_index_to_disk(self):
        """Store self.index to disk"""
        import scipy
        self.logger.info(f'Storing vectors to {self.index_abspath}')
        self.logger.info(f'Storing indices from {self.get_file_from_workspace(self.indices_filename)}')

        scipy.sparse.save_npz(
            self.index_abspath, scipy.sparse.vstack(self.index.values())
        )

        with open(self.get_file_from_workspace(self.indices_filename), 'wb') as f:
            np.save(f, list(self.index.keys()))

    def _load_index_from_disk(self):
        """Load self.index from disk"""
        import scipy
        self.logger.info(f'Loading vectors from {self.index_abspath}')
        self.logger.info(f'Loading indices from {self.get_file_from_workspace(self.indices_filename)}')

        vectors = scipy.sparse.load_npz(self.index_abspath)

        with open(self.get_file_from_workspace(self.indices_filename), 'rb') as f:
            indices = np.load(f)

        self.index = {ind: vec for ind, vec in zip(indices, vectors)}



import argparse
import asyncio
import copy
import json
from argparse import Namespace
from contextlib import ExitStack
from typing import Optional, Sequence, Dict

from pkg_resources import resource_filename

from .... import __default_host__
from ....enums import replace_enum_to_str
from ....importer import ImportExtensions
from ....jaml.helper import complete_path
from ....logging import JinaLogger


class DaemonClient:
    """
    Jina Daemon client.

    :param host: the host address of ``jinad`` instance
    :param port: the port number of ``jinad`` instance
    :param logger: Jinalogger to log information.
    :param timeout: stop waiting for a response after a given number of seconds with the timeout parameter.
    """

    kind = 'pea'  # select from pea/pod, TODO: enum

    def __init__(
        self,
        host: str,
        port: int,
        logger: 'JinaLogger' = None,
        timeout: int = 5,
        **kwargs,
    ):
        self.logger = logger or JinaLogger(host)
        self.timeout = timeout
        # for now it is http. but it can be https or unix socket or fd
        # TODO: for https, the jinad server would need a tls certificate.
        # no changes would be required in terms of how the api gets invoked,
        # as requests does ssl verfication. we'd need to add some exception handling logic though
        base_url = f'{host}:{port}'
        rest_url = f'http://{base_url}'
        self.alive_api = f'{rest_url}/'
        self.upload_api = f'{rest_url}/workspaces'
        self.upload_api_arg = 'files'  # this is defined in Daemon API upload interface
        if self.kind == 'pea':
            self.store_api = f'{rest_url}/peas'
        elif self.kind == 'pod':
            self.store_api = f'{rest_url}/pods'
        else:
            raise ValueError(f'{self.kind} is not supported')
        self.logstream_api = f'ws://{base_url}/logstream'

    @property
    def is_alive(self) -> bool:
        """
        Return True if ``jinad`` is alive at remote

        :return: True if ``jinad`` is alive at remote else false
        """
        with ImportExtensions(required=True):
            import requests

        try:
            r = requests.get(url=self.alive_api, timeout=self.timeout)
            return r.status_code == requests.codes.ok
        except requests.exceptions.RequestException as ex:
            self.logger.error(f'remote manager is not alive: {ex!r}')
            return False

    def get_status(self, identity: str) -> Dict:
        """Get status of the remote Pea / Pod

        :param identity: UUID string based identity for the Pea
        :raises: requests.exceptions.RequestException
        :return: json response of the remote Pea / Pod status
        :rtype: Dict
        """

        with ImportExtensions(required=True):
            import requests

        try:
            r = requests.get(url=f'{self.store_api}/{identity}', timeout=self.timeout)
            rj = r.json()
            if r.status_code == 200:
                return rj
            raise requests.exceptions.RequestException(rj)
        except requests.exceptions.RequestException as ex:
            self.logger.error(f'can\'t get status of {self.kind}: {ex!r}')

    def upload(
        self, dependencies: Sequence[str], workspace_id: Optional[str] = None
    ) -> str:
        """Upload local file dependencies to remote server by extracting from the pea_args

        :param dependencies: file dependencies
        :type dependencies: Sequence[str]
        :param workspace_id: Workspace to which the files will get uploaded, defaults to None
        :type workspace_id: str
        :raises: requests.exceptions.RequestException
        :return: json response for upload
        :rtype: str
        """
        import requests

        with ExitStack() as file_stack:
            files = [
                (
                    self.upload_api_arg,
                    file_stack.enter_context(open(complete_path(f), 'rb')),
                )
                for f in dependencies
            ]

            if files:
                try:
                    self.logger.info(f'uploading {len(files)} file(s): {dependencies}')
                    r = requests.post(
                        url=self.upload_api,
                        files=files,
                        data={'workspace_id': workspace_id} if workspace_id else None,
                        timeout=self.timeout,
                    )
                    rj = r.json()
                    if r.status_code == 201:
                        return rj
                    else:
                        raise requests.exceptions.RequestException(rj)
                except requests.exceptions.RequestException as ex:
                    self.logger.error(f'fail to upload as {ex!r}')

    def create(self, args: 'Namespace') -> Optional[str]:
        """Create a remote Pea / Pod

        :param args: the arguments for remote Pea
        :type args: Namespace
        :raises: requests.exceptions.RequestException
        :return: the identity of the spawned Pea / Pod
        :rtype: Optional[str]
        """

        with ImportExtensions(required=True):
            import requests

        try:
            payload = replace_enum_to_str(vars(self._mask_args(args)))
            # set timeout to None if args.timeout_ready is -1 (wait forever)
            r = requests.post(
                url=self.store_api,
                json=payload,
                timeout=args.timeout_ready if args.timeout_ready != -1 else None,
            )
            rj = r.json()
            if r.status_code == 201:
                return rj
            elif r.status_code == 400:
                # known internal error
                rj_body = '\n'.join(j for j in rj['body'])
                self.logger.error(f'{rj["detail"]}\n{rj_body}')
            elif r.status_code == 422:
                self.logger.error(
                    'your payload is not correct, please follow the error message and double check'
                )
            raise requests.exceptions.RequestException(rj)
        except requests.exceptions.RequestException as ex:
            self.logger.error(f'fail to create as {ex!r}')

    async def logstream(self, workspace_id: str, log_id: str):
        """Websocket log stream from remote Pea / Pod

        :param workspace_id: the identity of the workspace
        :type workspace_id: str
        :param log_id:  the identity of that Pea / Pod
        :type log_id: str
        """
        with ImportExtensions(required=True):
            import websockets

        remote_log_config = resource_filename(
            'jina', '/'.join(('resources', 'logging.remote.yml'))
        )
        all_remote_loggers = {}
        try:
            async with websockets.connect(
                f'{self.logstream_api}/{workspace_id}/{log_id}'
            ) as websocket:
                async for log_line in websocket:
                    try:
                        ll = json.loads(log_line)
                        name = ll['name']
                        if name not in all_remote_loggers:
                            all_remote_loggers[name] = JinaLogger(
                                context=ll['host'], log_config=remote_log_config
                            )

                        all_remote_loggers[name].info(
                            '{host} {name} {type} {message}'.format_map(ll)
                        )
                    except json.decoder.JSONDecodeError:
                        continue
        except websockets.exceptions.ConnectionClosedOK:
            self.logger.warning(f'log streaming is disconnected')
        except websockets.exceptions.WebSocketException as e:
            self.logger.error(
                f'log streaming is disabled, you won\'t see logs on the remote\n Reason: {e!r}'
            )
        except asyncio.CancelledError:
            self.logger.info(f'log streaming is cancelled')
        finally:
            for l in all_remote_loggers.values():
                l.close()

    def delete(self, remote_id: str, **kwargs) -> bool:
        """
        Delete a remote pea/pod

        :param remote_id: the identity of that pea/pod
        :param kwargs: keyword arguments
        :return: True if the deletion is successful
        """
        with ImportExtensions(required=True):
            import requests

        try:
            url = f'{self.store_api}/{remote_id}'
            r = requests.delete(url=url, timeout=self.timeout)
            return r.status_code == 200
        except requests.exceptions.RequestException as ex:
            self.logger.error(f'fail to delete {remote_id} as {ex!r}')
            return False

    def _mask_args(self, args: 'argparse.Namespace'):
        _args = copy.deepcopy(args)

        # reset the runtime to ZEDRuntime or ContainerRuntime
        if _args.runtime_cls == 'JinadRuntime':
            if _args.uses.startswith('docker://'):
                _args.runtime_cls = 'ContainerRuntime'
            else:
                _args.runtime_cls = 'ZEDRuntime'

        # reset the host default host
        # TODO:/NOTE this prevents jumping from remote to another remote (Han: 2021.1.17)
        _args.host = __default_host__

        _args.log_config = ''  # do not use local log_config
        _args.upload_files = []  # reset upload files
        _args.noblock_on_start = False  # wait until start success

        changes = []
        for k, v in vars(_args).items():
            if v != getattr(args, k):
                changes.append(f'{k:>30s}: {str(getattr(args, k)):30s} -> {str(v):30s}')
        if changes:
            changes = [
                'note the following arguments have been masked or altered for remote purpose:'
            ] + changes
            self.logger.warning('\n'.join(changes))

        return _args


class PeaDaemonClient(DaemonClient):
    """Pea API, we might have different endpoints for peas & pods later"""

    kind = 'pea'


class PodDaemonClient(DaemonClient):
    """Pod API, we might have different endpoints for peas & pods later"""

    kind = 'pod'



from typing import Iterable

from .document import DocumentSet

if False:
    from ..document import Document


class ChunkSet(DocumentSet):
    """
    :class:`ChunkSet` inherits from :class:`DocumentSet`.
    It's a subset of Documents.

    :param docs_proto: Set of sub-documents (i.e chunks) of `reference_doc`
    :type docs_proto: :class:`Document`
    :param reference_doc: Reference :class:`Document` for the sub-documents
    :type reference_doc: :class:`Document`
    """

    def __init__(self, docs_proto, reference_doc: 'Document'):
        """
        Set constructor method.

        :param docs_proto: protobuf representation of the chunks
        :param reference_doc: parent document
        """
        super().__init__(docs_proto)
        self._ref_doc = reference_doc

    def append(self, document: 'Document', **kwargs) -> 'Document':
        """Add a sub-document (i.e chunk) to the current Document.

        :param document: Sub-document to be appended
        :type document: :class:`Document`
        :param kwargs: additional keyword arguments
        :return: the newly added sub-document in :class:`Document` view
        :rtype: :class:`Document`

        .. note::
            Comparing to :attr:`DocumentSet.append()`, this method adds more safeguard to
            make sure the added chunk is legit.
        """

        from ..document import Document

        c = self._docs_proto.add()
        c.CopyFrom(document.proto)
        chunk = Document(c)

        chunk.set_attrs(
            parent_id=self._ref_doc.id, granularity=self.granularity, **kwargs
        )

        if not chunk.mime_type:
            chunk.mime_type = self._ref_doc.mime_type
        chunk.update_content_hash()
        return chunk

    def extend(self, iterable: Iterable['Document']) -> None:
        """
        Extend the :class:`DocumentSet` by appending all the items from the iterable.

        :param iterable: the iterable of Documents to extend this set with
        """
        for doc in iterable:
            self.append(doc)
        num_siblings = len(self)
        for doc in self:
            doc.siblings = num_siblings

    @property
    def reference_doc(self) -> 'Document':
        """
        Get the document that :class:`ChunkSet` belongs to.

        :return: reference doc
        """
        return self._ref_doc

    @property
    def granularity(self) -> int:
        """
        Get granularity of all document in this set.

        :return: granularity
        """
        return self._ref_doc.granularity + 1

    @property
    def adjacency(self) -> int:
        """
        Get adjacency of all document in this set.

        :return: adjacency
        """
        return self._ref_doc.adjacency



"""Argparser module for remote runtime"""
from ...helper import add_arg_group
from .... import __default_host__
from ....enums import CompressAlgo
from .... import helper


def mixin_remote_parser(parser):
    """Add the options for remote expose
    :param parser: the parser
    """
    gp = add_arg_group(parser, title='Expose')

    gp.add_argument(
        '--host',
        type=str,
        default=__default_host__,
        help=f'The host address of the runtime, by default it is {__default_host__}.',
    )

    gp.add_argument(
        '--port-expose',
        type=int,
        default=helper.random_port(),
        help='The port of the host exposed to the public',
    )


def mixin_grpc_parser(parser=None):
    """Add the options for gRPC
    :param parser: the parser
    """
    gp = add_arg_group(parser, title='GRPC/REST')

    gp.add_argument(
        '--max-message-size',
        type=int,
        default=-1,
        help='The maximum send and receive size for gRPC server in bytes, -1 means unlimited',
    )
    gp.add_argument(
        '--proxy',
        action='store_true',
        default=False,
        help='If set, respect the http_proxy and https_proxy environment variables. '
        'otherwise, it will unset these proxy variables before start. '
        'gRPC seems to prefer no proxy',
    )
    gp.add_argument(
        '--prefetch',
        type=int,
        default=50,
        help='The number of pre-fetched requests from the client',
    )
    gp.add_argument(
        '--prefetch-on-recv',
        type=int,
        default=1,
        help='The number of additional requests to fetch on every receive',
    )
    gp.add_argument(
        '--restful',
        '--rest-api',
        action='store_true',
        default=False,
        help='If set, use RESTful interface instead of gRPC as the main interface',
    )
    gp.add_argument(
        '--compress',
        type=CompressAlgo.from_string,
        choices=list(CompressAlgo),
        default=CompressAlgo.LZ4,
        help='''
The compress algorithm used over the entire Flow.

Note that this is not necessarily effective, it depends on the settings of `--compress-lwm` and `compress-hwm`''',
    )
    gp.add_argument(
        '--compress-min-bytes',
        type=int,
        default=1024,
        help='The original message size must be larger than this number to trigger the compress algorithm, '
        '-1 means disable compression.',
    )
    gp.add_argument(
        '--compress-min-ratio',
        type=float,
        default=1.1,
        help='The compression ratio (uncompressed_size/compressed_size) must be higher than this number '
        'to trigger the compress algorithm.',
    )



import argparse
from typing import Dict, Any

from ..base import VersionedYAMLParser
from ....enums import PodRoleType
from ....flow import Flow
from ....helper import expand_env_var, ArgNamespace
from ....parsers import set_pod_parser, set_gateway_parser


def _get_taboo():
    """
    :return: set of keys that should not be dumped
    """
    return {k.dest for k in set_pod_parser()._actions if k.help == argparse.SUPPRESS}


class V1Parser(VersionedYAMLParser):
    """V1Parser introduces new syntax and features:

        - It has a top-level field ``version``
        - ``pods`` is now a List of Dict (rather than a Dict as prev.)
        - ``name`` is now optional
        - new field ``method`` can be used to specify how to add this Pod into the Flow, availables are:
            - ``add``: (default) equal to `Flow.add(...)`
            - ``needs``: (default) equal to `Flow.needs(...)`
            - ``inspect``: (default) equal to `Flow.inspect(...)`

    An example V1 YAML config can be found below:
        .. highlight:: yaml
        .. code-block:: yaml

            !Flow
            version: '1.0'
            pods:
              - name: pod0  # notice the change here, name is now an attribute
                method: add  # by default method is always add, available: add, needs, inspect
                uses: _pass
                needs: gateway
              - name: pod1  # notice the change here, name is now an attribute
                method: add  # by default method is always add, available: add, needs, inspect
                uses: _pass
                needs: gateway
              - method: inspect  # add an inspect node on pod1
              - method: needs  # let's try something new in Flow YAML v1: needs
                needs: [pod1, pod0]


    """

    version = '1'  # the version number this parser designed for

    def parse(self, cls: type, data: Dict) -> 'Flow':
        """
        :param cls: the class registered for dumping/loading
        :param data: flow yaml file loaded as python dict
        :return: the Flow YAML parser given the syntax version number
        """
        envs = data.get('env', {})  # type: Dict[str, str]
        p = data.get('with', {})  # type: Dict[str, Any]
        a = p.pop('args') if 'args' in p else ()
        k = p.pop('kwargs') if 'kwargs' in p else {}
        # maybe there are some hanging kwargs in "parameters"
        tmp_a = (expand_env_var(v) for v in a)
        tmp_p = {kk: expand_env_var(vv) for kk, vv in {**k, **p}.items()}
        obj = cls(*tmp_a, env=envs, **tmp_p)

        pp = data.get('pods', [])
        for pods in pp:
            p_pod_attr = {kk: expand_env_var(vv) for kk, vv in pods.items()}
            # in v1 YAML, flow is an optional argument
            if p_pod_attr.get('name', None) != 'gateway':
                # ignore gateway when reading, it will be added during build()
                method = p_pod_attr.get('method', 'add')
                # support methods: add, needs, inspect
                getattr(obj, method)(**p_pod_attr, copy_flow=False)
        return obj

    def dump(self, data: 'Flow') -> Dict:
        """
        :param data: versioned flow object
        :return: the dictionary given a versioned flow object
        """
        r = {}
        if data._version:
            r['version'] = data._version

        if data._env:
            r['env'] = data._env

        if data._kwargs:
            r['with'] = data._kwargs

        if data._pod_nodes:
            r['pods'] = []

        last_name = 'gateway'
        for k, v in data._pod_nodes.items():
            if k == 'gateway':
                continue
            kwargs = {}
            # only add "needs" when the value is not the last pod name
            if list(v.needs) != [last_name]:
                kwargs = {'needs': list(v.needs)}

            # get nondefault kwargs
            parser = set_pod_parser()
            if v.role == PodRoleType.GATEWAY:
                parser = set_gateway_parser()

            non_default_kw = ArgNamespace.get_non_defaults_args(v.args, parser)

            kwargs.update(non_default_kw)

            for t in _get_taboo():
                if t in kwargs:
                    kwargs.pop(t)
            last_name = kwargs['name']
            r['pods'].append(kwargs)
        return r



import time
from collections import defaultdict
from typing import Dict, List

import zmq

from .base import ZMQRuntime
from ...zmq import ZmqStreamlet
from .... import Message
from .... import Request
from ....enums import OnErrorStrategy
from ....excepts import (
    NoExplicitMessage,
    ExecutorFailToLoad,
    MemoryOverHighWatermark,
    ChainedPodException,
    BadConfigSource,
    RuntimeTerminated,
)
from ....executors import BaseExecutor
from ....helper import random_identity
from ....logging.profile import used_memory, TimeDict
from ....proto import jina_pb2


class ZEDRuntime(ZMQRuntime):
    """Runtime procedure leveraging :class:`ZmqStreamlet` for Executor, Driver."""

    def run_forever(self):
        """Start the `ZmqStreamlet`."""
        self._zmqlet.start(self._msg_callback)

    def setup(self):
        """Initialize private parameters and execute private loading functions."""
        self._id = random_identity()
        self._last_active_time = time.perf_counter()
        self._last_dump_time = time.perf_counter()
        self._last_load_time = time.perf_counter()

        self._timer = TimeDict()

        self._request = None
        self._message = None

        # all pending messages collected so far, key is the request id
        self._pending_msgs = defaultdict(list)  # type: Dict[str, List['Message']]
        self._partial_requests = None
        self._partial_messages = None

        self._load_zmqlet()
        self._load_plugins()
        self._load_executor()

    def teardown(self):
        """Close the `ZmqStreamlet` and `Executor`."""
        self._zmqlet.close()
        self._executor.close()
        super().teardown()

    #: Private methods required by :meth:`setup`

    def _load_zmqlet(self):
        """Load ZMQStreamlet to this runtime."""
        # important: fix zmqstreamlet ctrl address to replace the the ctrl address generated in the main
        # process/thread
        self._zmqlet = ZmqStreamlet(
            self.args, logger=self.logger, ctrl_addr=self.ctrl_addr
        )

    def _load_executor(self):
        """Load the executor to this runtime, specified by ``uses`` CLI argument."""
        try:
            self._executor = BaseExecutor.load_config(
                self.args.uses, pea_id=self.args.pea_id, read_only=self.args.read_only
            )
            self._executor.attach(runtime=self)
        except BadConfigSource as ex:
            self.logger.error(
                f'fail to load config from {self.args.uses}, if you are using docker image for --uses, '
                f'please use "docker://YOUR_IMAGE_NAME"'
            )
            raise ExecutorFailToLoad from ex
        except FileNotFoundError as ex:
            self.logger.error(f'fail to load file dependency')
            raise ExecutorFailToLoad from ex
        except Exception as ex:
            self.logger.critical(f'can not load the executor from {self.args.uses}')
            raise ExecutorFailToLoad from ex

    def _load_plugins(self):
        """Load the plugins if needed necessary to load executors."""
        if self.args.py_modules:
            from ....importer import PathImporter

            PathImporter.add_modules(*self.args.py_modules)

    #: Private methods required by :meth:`teardown`

    def _save_executor(self):
        """Save the contained executor according to the `dump_interval` parameter."""
        if (time.perf_counter() - self._last_dump_time) > self.args.dump_interval > 0:
            self._executor.save()
            self._last_dump_time = time.perf_counter()

    def _reload_executor(self):
        if (time.perf_counter() - self._last_load_time) > self.args.load_interval > 0:
            self._load_executor()
            self._last_load_time = time.perf_counter()

    def _check_memory_watermark(self):
        """Check the memory watermark."""
        if used_memory() > self.args.memory_hwm > 0:
            raise MemoryOverHighWatermark

    #: Private methods required by run_forever
    def _pre_hook(self, msg: 'Message') -> 'ZEDRuntime':
        """
        Pre-hook function, what to do after first receiving the message.

        :param msg: received message
        :return: `ZEDRuntime`
        """
        msg.add_route(self.name, self._id)
        self._reload_executor()
        self._request = msg.request
        self._message = msg

        part_str = ''
        if self.expect_parts > 1:
            req_id = msg.envelope.request_id
            self._pending_msgs[req_id].append(msg)
            self._partial_messages = self._pending_msgs[req_id]
            self._partial_requests = [v.request for v in self._partial_messages]
            part_str = f'({len(self.partial_requests)}/{self.expect_parts} parts)'

        self.logger.info(
            f'recv {msg.envelope.request_type} {part_str} from {msg.colored_route}'
        )
        return self

    def _post_hook(self, msg: 'Message') -> 'ZEDRuntime':
        """
        Post-hook function, what to do before handing out the message.

        :param msg: received message
        :return: `ZEDRuntime`
        """
        self._last_active_time = time.perf_counter()
        self._save_executor()
        self._zmqlet.print_stats()
        self._check_memory_watermark()

        if self.expect_parts > 1:
            msgs = self._pending_msgs.pop(msg.envelope.request_id)
            msg.merge_envelope_from(msgs)

        msg.update_timestamp()
        return self

    def _handle(self, msg: 'Message') -> 'ZEDRuntime':
        """Register the current message to this pea, so that all message-related properties are up-to-date, including
        :attr:`request`, :attr:`prev_requests`, :attr:`message`, :attr:`prev_messages`. And then call the executor to handle
        this message if its envelope's  status is not ERROR, else skip handling of message.

        :param msg: the message received
        :return: ZEDRuntime procedure.
        """
        if self.expect_parts > 1 and self.expect_parts > len(self.partial_requests):
            # NOTE: reduce priority is higher than chain exception
            # otherwise a reducer will lose its function when eailier pods raise exception
            raise NoExplicitMessage

        if (
            msg.envelope.status.code != jina_pb2.StatusProto.ERROR
            or self.args.on_error_strategy < OnErrorStrategy.SKIP_HANDLE
        ):
            self._executor(self.request_type)
        else:
            raise ChainedPodException
        return self

    def _callback(self, msg: 'Message'):
        self.is_post_hook_done = False  #: if the post_hook is called
        self._pre_hook(msg)._handle(msg)._post_hook(msg)
        self.is_post_hook_done = True
        return msg

    def _msg_callback(self, msg: 'Message') -> None:
        """
        Callback function after receiving the message
        When nothing is returned then nothing is send out via :attr:`zmqlet.sock_out`.

        :param msg: received message
        """
        try:
            # notice how executor related exceptions are handled here
            # generally unless executor throws an OSError, the exception are caught and solved inplace
            self._zmqlet.send_message(self._callback(msg))
        except RuntimeTerminated:
            # this is the proper way to end when a terminate signal is sent
            self._zmqlet.send_message(msg)
            self._zmqlet.close()
        except (SystemError, zmq.error.ZMQError, KeyboardInterrupt) as ex:
            # save executor
            self.logger.info(f'{ex!r} causes the breaking from the event loop')
            self._zmqlet.send_message(msg)
            self._zmqlet.close()
        except MemoryOverHighWatermark:
            self.logger.critical(
                f'memory usage {used_memory()} GB is above the high-watermark: {self.args.memory_hwm} GB'
            )
        except NoExplicitMessage:
            # silent and do not propagate message anymore
            # 1. wait partial message to be finished
            # 2. dealer send a control message and no need to go on
            pass
        except (RuntimeError, Exception, ChainedPodException) as ex:
            # general runtime error and nothing serious, we simply mark the message to error and pass on
            if not self.is_post_hook_done:
                self._post_hook(msg)

            if self.args.on_error_strategy == OnErrorStrategy.THROW_EARLY:
                raise
            if isinstance(ex, ChainedPodException):
                msg.add_exception()
                self.logger.error(
                    f'{ex!r}'
                    + f'\n add "--quiet-error" to suppress the exception details'
                    if not self.args.quiet_error
                    else '',
                    exc_info=not self.args.quiet_error,
                )
            else:
                msg.add_exception(ex, executor=getattr(self, '_executor'))
                self.logger.error(
                    f'{ex!r}'
                    + f'\n add "--quiet-error" to suppress the exception details'
                    if not self.args.quiet_error
                    else '',
                    exc_info=not self.args.quiet_error,
                )

            self._zmqlet.send_message(msg)

    #: Some class-specific properties

    @property
    def is_idle(self) -> bool:
        """
        Return ``True`` when current time is ``max_idle_time`` seconds late than the last active time

        :return: True if idle else false.
        """
        return (time.perf_counter() - self._last_active_time) > self.args.max_idle_time

    @property
    def request(self) -> 'Request':
        """
        Get the current request body inside the protobuf message

        :return: :class:`ZEDRuntime` request
        """
        return self._request

    @property
    def message(self) -> 'Message':
        """
        Get the current protobuf message to be processed

        :return: :class:`ZEDRuntime` message
        """
        return self._message

    @property
    def request_type(self) -> str:
        """
        Get the type of message being processed

        :return: request type
        """
        return self._message.envelope.request_type

    @property
    def expect_parts(self) -> int:
        """
        The expected number of partial messages before trigger :meth:`handle`

        :return: expected number of partial messages
        """
        return self.args.num_part if self.message.is_data_request else 1

    @property
    def partial_requests(self) -> List['Request']:
        """
        The collected partial requests under the current ``request_id``

        :return: collected partial requests
        """
        return self._partial_requests

    @property
    def partial_messages(self) -> List['Message']:
        """
        The collected partial messages under the current ``request_id`` "
        :return: collected partial messages

        """
        return self._partial_messages



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Dict, Any, Iterable

from ...types.querylang.queryset.lookup import Q
from .. import QuerySetReader, BaseRecursiveDriver, ContextAwareRecursiveMixin

if False:
    from ...types.sets import DocumentSet


class FilterQL(QuerySetReader, ContextAwareRecursiveMixin, BaseRecursiveDriver):
    """Filters incoming `docs` by evaluating a series of `lookup rules`.

    This is often useful when the proceeding Pods require only a signal, not the full message.

    Example ::
    - !FilterQL
        with:
            lookups: {modality: mode2}
    - !EncodeDriver
        with:
            method: encode

    ensures that the EncodeDriver will only get documents which modality field value is `mode2` by filtering
    those documents at the specific levels that do not comply with this condition

    :param lookups: (dict) a dictionary where keys are interpreted by ``:class:`LookupLeaf`` to form a
    an evaluation function. For instance, a dictionary ``{ modality__in: [mode1, mode2] }``, would create
    an evaluation function that will check if the field `modality` is found in `[mode1, mode2]`
    :param args: additional positional arguments which are just used for the parent initialization
    :param kwargs: additional key value arguments which are just used for the parent initialization
    """

    def __init__(self, lookups: Dict[str, Any], *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._lookups = lookups

    def _apply_all(
        self, doc_sequences: Iterable['DocumentSet'], *args, **kwargs
    ) -> None:
        for docs in doc_sequences:
            if self.lookups:
                _lookups = Q(**self.lookups)
                miss_idx = []
                for idx, doc in enumerate(docs):
                    if not _lookups.evaluate(doc):
                        miss_idx.append(idx)

                # delete non-exit matches in reverse
                for j in reversed(miss_idx):
                    del docs[j]



from typing import Optional
import numpy as np

from jina.executors.encoders.frameworks import BasePaddleEncoder
from jina.executors.decorators import batching, as_ndarray


class TextPaddlehubEncoder(BasePaddleEncoder):
    """
    Encode an array of string in size `B` into an ndarray in size `B x D`

    The ndarray potentially is BatchSize x (Channel x Height x Width)

    Internally, :class:`TextPaddlehubEncoder` wraps the Ernie module from paddlehub.
    https://github.com/PaddlePaddle/PaddleHub
    For models' details refer to
    https://www.paddlepaddle.org.cn/hublist?filter=en_category&value=SemanticModel

    :param model_name: the name of the model. Supported models include
        ``ernie``, ``ernie_tiny``, ``ernie_v2_eng_base``, ``ernie_v2_eng_large``,
        ``bert_chinese_L-12_H-768_A-12``, ``bert_multi_cased_L-12_H-768_A-12``,
        ``bert_multi_uncased_L-12_H-768_A-12``, ``bert_uncased_L-12_H-768_A-12``,
        ``bert_uncased_L-24_H-1024_A-16``, ``chinese-bert-wwm``,
        ``chinese-bert-wwm-ext``, ``chinese-electra-base``,
        ``chinese-electra-small``, ``chinese-roberta-wwm-ext``,
        ``chinese-roberta-wwm-ext-large``, ``rbt3``, ``rbtl3``
    :param args:  Additional positional arguments
    :param kwargs: Additional keyword arguments
    """

    def __init__(self, model_name: Optional[str] = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model_name = model_name or 'ernie_tiny'

    def post_init(self):
        """Load PaddleHub model"""
        import paddlehub as hub
        self.model = hub.Module(name=self.model_name)

    @batching
    @as_ndarray
    def encode(self, content: 'np.ndarray', *args, **kwargs) -> 'np.ndarray':
        """
        Encode ``Document`` content from an array of string in size `B` into a ndarray in size `B x D`.

        :param content: a 1d array of string type in size `B`
        :return: an ndarray in size `B x D`
        """
        results = []
        _raw_results = self.model.get_embedding(
            np.atleast_2d(content).reshape(-1, 1).tolist(), use_gpu=self.on_gpu)
        for emb in _raw_results:
            _pooled_feature, _seq_feature = emb
            results.append(_pooled_feature)
        return np.array(results)



"""Module wrapping interactions with the local images."""

import os
import pkgutil
from typing import Dict, Any, List, Optional

from pkg_resources import parse_version
from setuptools import find_packages

from ..helper import credentials_file
from ...excepts import HubLoginRequired
from ...helper import colored
from ...jaml import JAML
from ...logging import default_logger
from ... import __version__

_header_attrs = ['bold', 'underline']


def _load_local_hub_manifest():
    from ..checker import safe_url_name

    namespace = 'jina.hub'
    try:
        path = os.path.dirname(pkgutil.get_loader(namespace).path)
    except AttributeError:
        default_logger.warning(
            'local Hub is not initialized, '
            'try "git submodule update --init" if you are in dev mode'
        )
        return {}

    def _add_hub():
        m_yml = f'{info.module_finder.path}/manifest.yml'
        if info.ispkg and os.path.exists(m_yml):
            try:
                with open(m_yml) as fp:
                    m = JAML.load(fp)
                    m['source_path'] = info.module_finder.path
                    exec_name = safe_url_name(
                        f'jinahub/' + f'{m["type"]}.{m["kind"]}.{m["name"]}'
                    )
                    tag = exec_name + f':{m["version"]}-{__version__}'
                    m['image_tag'] = tag
                    hub_images[m['name']] = m
            except:
                pass  #: intentionally suppress the error, as some dirs do not fit to a valid hub structure

    hub_images = {}

    for info in pkgutil.iter_modules([path]):
        _add_hub()

    for pkg in find_packages(path):
        pkgpath = path + '/' + pkg.replace('.', '/')
        for info in pkgutil.iter_modules([pkgpath]):
            _add_hub()

    # filter
    return hub_images


def _list_local(logger) -> Optional[Dict[str, Any]]:
    """
    List the locally-available images.

    :param logger: the logger object with which to print
    :return: the list of manifests (if found)
    """
    manifests = _load_local_hub_manifest()
    if manifests:
        tb = _make_hub_table(manifests.values())
        logger.info('\n'.join(tb))
    return manifests


def _fetch_access_token(logger) -> Optional[str]:
    """
    Fetch github access token from credentials file, return as a request header

    :param logger: the logger object with which to print
    :return: local github access token (if found)
    """
    logger.info('fetching github access token...')

    if not credentials_file().is_file():
        raise HubLoginRequired(
            f'❌ User not logged in. please login using command: {colored("jina hub login", attrs=["bold"])}'
        )
    try:
        with open(credentials_file(), 'r') as cf:
            cred_yml = JAML.load(cf)
            access_token = cred_yml['access_token']
            return access_token
    except Exception:
        raise HubLoginRequired(
            f'❌ Invalid access file. '
            f'please re-login using command: {colored("jina hub login", attrs=["bold"])}'
        )


def _make_hub_table_with_local(images: Dict, local_images: Dict) -> List[str]:
    """
    Pretty print remote and local hub images

    :param images: remote images as a dict
    :param local_images: local images as a dict
    :return: table content to be pretty printed as a list
    """

    info_table = [
        f'found {len(images)} matched hub images',
        '{:<50s}{:<25s}{:<30s}{:<25s}{:<30s}{:<50s}'.format(
            colored('Name', attrs=_header_attrs),
            colored('Kind', attrs=_header_attrs),
            colored('Version', attrs=_header_attrs),
            colored('Local', attrs=_header_attrs),
            colored('Jina Version', attrs=_header_attrs),
            colored('Description', attrs=_header_attrs),
        ),
    ]
    images = sorted(images, key=lambda k: k['name'].lower())
    for image in images:
        image_name = image.get('name', '')
        kind = image.get('kind', '')
        ver = image.get('version', '')
        jina_ver = image.get('jina-version', '')
        desc = image.get('description', '')[:60].strip() + '...'
        if image_name and ver and desc:
            local_ver = ''
            color = 'white'
            if image_name in local_images:
                local_ver = local_images[image_name].get('version', '')
                _v1, _v2 = parse_version(ver), parse_version(local_ver)
                if _v1 > _v2:
                    color = 'red'
                elif _v1 == _v2:
                    color = 'green'
                else:
                    color = 'yellow'
            info_table.append(
                f'{colored(image_name, color="yellow", attrs="bold"):<50s}'
                f'{colored(kind, color="yellow"):<25s}'
                f'{colored(ver, color="green"):<25s}'
                f'{colored(local_ver, color=color):<25s}'
                f'{colored(jina_ver, color="green"):<25s}'
                f'{desc:<30s}'
            )
    return info_table


def _make_hub_table(images) -> List[str]:
    """
    Pretty print local hub images

    :param images: local images as a dict
    :return: table content to be pretty printed as a list
    """
    info_table = [
        f'found {len(images)} matched hub images',
        '{:<50s}{:<25s}{:<25s}{:<30s}'.format(
            colored('Name', attrs=_header_attrs),
            colored('Kind', attrs=_header_attrs),
            colored('Version', attrs=_header_attrs),
            colored('Description', attrs=_header_attrs),
        ),
    ]
    images = sorted(images, key=lambda k: k['name'].lower())
    for image in images:
        image_name = image.get('name', '')
        kind = image.get('kind', '')
        ver = image.get('version', '')
        desc = image.get('description', '')[:60].strip() + '...'
        if image_name and ver and desc:
            info_table.append(
                f'{colored(image_name, color="yellow", attrs="bold"):<50s}'
                f'{colored(kind, color="yellow"):<25s}'
                f'{colored(ver, color="green"):<25s}'
                f'{desc:<30s}'
            )
    return info_table



from . import Message
from ..request import Request
from ...proto import jina_pb2

_available_commands = dict(
    jina_pb2.RequestProto.ControlRequestProto.DESCRIPTOR.enum_values_by_name
)

__all__ = ['ControlMessage']


class ControlMessage(Message):
    """
    Class of the protobuf message.

    :param command: Command with string content. (e.g. 'IDLE', 'CANCEL', 'TERMINATE', 'STATUS')
    :param pod_name: Name of the current pod.
    :param identity: The identity of the current pod
    :param args: Additional positional arguments which are just used for the parent initialization
    :param kwargs: Additional keyword arguments which are just used for the parent initialization
    """

    def __init__(
        self, command: str, pod_name: str = 'ctl', identity: str = '', *args, **kwargs
    ):
        req = Request(jina_pb2.RequestProto())
        if command in _available_commands:
            req.control.command = getattr(
                jina_pb2.RequestProto.ControlRequestProto, command
            )
        else:
            raise ValueError(
                f'command "{command}" is not supported, must be one of {_available_commands}'
            )
        super().__init__(
            None, req, pod_name=pod_name, identity=identity, *args, **kwargs
        )
        req.request_type = 'control'



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Dict, List

from ..jaml import JAML

_defaults = {}


def get_default_reqs(cls_mro: List[type]) -> Dict:
    """Get a copy of default meta variables

    :param cls_mro: the MRO inherited order followed.
    """
    import copy

    global _defaults

    for cls in cls_mro:
        try:
            if cls.__name__ not in _defaults:
                from pkg_resources import resource_stream

                with resource_stream(
                    'jina',
                    '/'.join(('resources', f'executors.requests.{cls.__name__}.yml')),
                ) as fp:
                    _defaults[cls.__name__] = JAML.load(
                        fp
                    )  # do not expand variables at here, i.e. DO NOT USE expand_dict(yaml.load(fp))

            if cls.__name__ != cls_mro[0].__name__:
                from ..logging import default_logger

                default_logger.debug(
                    f'"requests.on" setting of {cls_mro[0]} fallback to general {cls} setting, '
                    f'because you did not specify {cls_mro[0]}'
                )
            return copy.deepcopy(_defaults[cls.__name__])
        except FileNotFoundError:
            pass

    raise ValueError(
        f'not able to find any default settings along this chain {cls_mro!r}'
    )



from typing import Iterable


class Traversable:
    """
    Helper class to expose common interface to the traversal logic of the BaseExecutable Driver.
    It is important to note that it checks the matching structure of `docs` and `groundtruths`. It is important while
    traversing to ensure that then the driver can be applied at a comparable level of granularity and adjacency.
    This does not imply that you can't compare at the end a document with 10 matches with a groundtruth with 20 matches

    :param doc: Target `Document`.
    :param groundtruth: The :class:`Document` with desired state.
    """

    @property
    def matches(self) -> Iterable['Traversable']:
        """Get the pairs between matches and Groundtruth."""

    @property
    def chunks(self) -> Iterable['Traversable']:
        """Get the pairs between chunks and Groundtruth."""



import pprint
from typing import Dict

from google.protobuf.json_format import MessageToJson, MessageToDict

from ..helper import typename
from ..proto import jina_pb2


class ProtoTypeMixin:
    """Mixin class of `ProtoType`."""

    def json(self) -> str:
        """Return the object in JSON string

        :return: JSON string of the object
        """
        return MessageToJson(self._pb_body)

    def dict(self) -> Dict:
        """Return the object in Python dictionary

        :return: dict representation of the object
        """
        return MessageToDict(self._pb_body)

    @property
    def proto(self) -> 'jina_pb2._reflection.GeneratedProtocolMessageType':
        """Return the underlying Protobuf object

        :return: Protobuf representation of the object
        """
        return self._pb_body

    @property
    def binary_str(self) -> bytes:
        """Return the serialized the message to a string.

        :return: binary string representation of the object
        """
        return self._pb_body.SerializeToString()

    def __getattr__(self, name: str):
        return getattr(self._pb_body, name)

    def __str__(self):
        return str(self._build_content_dict())

    def __repr__(self):
        content = ' '.join(f'{k}={v}' for k, v in self._build_content_dict().items())
        content += f' at {id(self)}'
        content = content.strip()
        return f'<{typename(self)} {content}>'

    def _build_content_dict(self):
        """Helper method for __str__ and __repr__

        :return: the dict representation for the object
        """
        content = self.dict()
        if hasattr(self, '_attributes_in_str') and isinstance(
            self._attributes_in_str, list
        ):
            content = {k: content[k] for k in self._attributes_in_str}
        return content



import asyncio
from abc import ABC

import zmq.asyncio

from ..zmq.base import ZMQRuntime
from ...zmq import _init_socket, recv_message_async, send_message_async
from ....enums import SocketType
from ....proto import jina_pb2


class AsyncZMQRuntime(ZMQRuntime):
    """
    Runtime procedure in the async manners.

    Base class of :class:`AsyncNewLoopRuntime`.
    """

    def run_forever(self):
        """Running method to block the main thread."""
        asyncio.run(self._loop_body())

    async def async_cancel(self):
        """An async method to cancel."""
        raise NotImplementedError

    async def async_run_forever(self):
        """The async method to run until it is stopped."""
        raise NotImplementedError

    async def _wait_for_cancel(self):
        """Do NOT override this method when inheriting from :class:`GatewayPea`"""
        with zmq.asyncio.Context() as ctx, _init_socket(
            ctx, self.ctrl_addr, None, SocketType.PAIR_BIND, use_ipc=True
        )[0] as sock:
            msg = await recv_message_async(sock)
            if msg.request.command == 'TERMINATE':
                msg.envelope.status.code = jina_pb2.StatusProto.SUCCESS
                await self.async_cancel()
                await send_message_async(sock, msg)

    async def _loop_body(self):
        """Do NOT override this method when inheriting from :class:`GatewayPea`"""
        try:
            await asyncio.gather(self.async_run_forever(), self._wait_for_cancel())
        except asyncio.CancelledError:
            self.logger.warning('received terminate ctrl message from main process')
        await self.async_cancel()


class AsyncNewLoopRuntime(AsyncZMQRuntime, ABC):
    """
    The runtime to start a new event loop.

    Base class for :class:`GRPCRuntime` and :class:`RESTRuntime`.
    """

    def run_forever(self):
        """
        Running method to block the main thread.

        Run the event loop until a Future is done.
        """
        self._loop.run_until_complete(self._loop_body())

    def setup(self):
        """Setup the event loop."""
        self._loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self._loop)
        self._loop.run_until_complete(self.async_setup())

    def teardown(self):
        """Stop and close the event loop."""
        self._loop.stop()
        self._loop.close()
        super().teardown()

    async def async_setup(self):
        """The async method to setup."""
        pass



import asyncio
from typing import Iterable, Any

from .....helper import random_identity, typename
from .....logging import JinaLogger
from .....logging.profile import TimeContext
from .....proto import jina_pb2_grpc
from .....types.message import Message
from .....types.request import Request

__all__ = ['AsyncPrefetchCall']


class AsyncPrefetchCall(jina_pb2_grpc.JinaRPCServicer):
    """JinaRPCServicer """

    def __init__(self, args, zmqlet):
        super().__init__()
        self.args = args
        self.zmqlet = zmqlet
        self.name = args.name or self.__class__.__name__
        self.logger = JinaLogger(self.name, **vars(args))
        self._id = random_identity()

    async def Call(self, request_iterator, context):
        """
        Async gRPC call.

        :param request_iterator: iterator of request.
        :param context: gRPC context:
        :yield: task
        """

        def handle(msg: 'Message') -> 'Request':
            """
            Add route into the `message` and return response of the message.

            :param msg: gRPC message.
            :return: message with route added.
            """
            msg.add_route(self.name, self._id)
            return msg.response

        async def prefetch_req(num_req, fetch_to):
            """
            Fetch and send request.

            :param num_req: number of requests
            :param fetch_to: the task list storing requests
            :return: False if append task to :param:`fetch_to` else False
            """
            for _ in range(num_req):
                try:
                    if hasattr(request_iterator, '__anext__'):
                        next_request = await request_iterator.__anext__()
                    elif hasattr(request_iterator, '__next__'):
                        # This code block will be executed for REST based invocations
                        next_request = next(request_iterator)
                    else:
                        raise TypeError(
                            f'{typename(request_iterator)} does not have `__anext__` or `__next__`'
                        )
                    asyncio.create_task(
                        self.zmqlet.send_message(
                            Message(None, next_request, 'gateway', **vars(self.args))
                        )
                    )
                    fetch_to.append(
                        asyncio.create_task(self.zmqlet.recv_message(callback=handle))
                    )
                except (StopIteration, StopAsyncIteration):
                    return True
            return False

        with TimeContext(f'prefetching {self.args.prefetch} requests', self.logger):
            self.logger.warning(
                'if this takes too long, you may want to take smaller "--prefetch" or '
                'ask client to reduce "--request-size"'
            )
            prefetch_task = []
            is_req_empty = await prefetch_req(self.args.prefetch, prefetch_task)
            if is_req_empty and not prefetch_task:
                self.logger.error(
                    'receive an empty stream from the client! '
                    'please check your client\'s inputs, '
                    'you can use "Client.check_input(inputs)"'
                )
                return

        # the total num requests < self.args.prefetch
        if is_req_empty:
            for r in asyncio.as_completed(prefetch_task):
                yield await r
        else:
            # if there are left over (`else` clause above is unnecessary for code but for better readability)
            onrecv_task = []
            # the following code "interleaves" prefetch_task and onrecv_task, when one dries, it switches to the other
            while prefetch_task:
                self.logger.info(
                    f'send: {self.zmqlet.msg_sent} '
                    f'recv: {self.zmqlet.msg_recv} '
                    f'pending: {self.zmqlet.msg_sent - self.zmqlet.msg_recv}'
                )
                onrecv_task.clear()
                for r in asyncio.as_completed(prefetch_task):
                    yield await r
                    if not is_req_empty:
                        is_req_empty = await prefetch_req(
                            self.args.prefetch_on_recv, onrecv_task
                        )

                # this list dries, clear it and feed it with on_recv_task
                prefetch_task.clear()
                prefetch_task = [j for j in onrecv_task]



from typing import Optional

from jina.executors.indexers.dump import import_metas
from jina.executors.indexers.keyvalue import BinaryPbWriterMixin
from jina.executors.indexers.query import BaseQueryIndexer


class BinaryPbQueryIndexer(BinaryPbWriterMixin, BaseQueryIndexer):
    """A write-once Key-value indexer."""

    def _load_dump(self, dump_path):
        """Load the dump at the path

        :param dump_path: the path of the dump"""
        ids, metas = import_metas(dump_path, str(self.pea_id))
        self._add(list(ids), list(metas))
        self.write_handler.flush()
        self.write_handler.close()
        self.handler_mutex = False
        self.is_handler_loaded = False
        del self.write_handler
        # warming up
        self._query('someid')

    def query(self, key: str, *args, **kwargs) -> Optional[bytes]:
        """Get a document by its id

        :param key: the id
        :param args: not used
        :param kwargs: not used
        :return: the bytes of the Document
        """
        return self._query(key)


class KeyValueQueryIndexer(BinaryPbQueryIndexer):
    """An alias"""



from typing import Tuple, Dict, List, Union

import numpy as np

from jina.executors.decorators import single
from jina.executors.segmenters import BaseSegmenter

from .helper import _move_channel_axis


class SlidingWindowImageCropper(BaseSegmenter):
    """
    :class:`SlidingWindowImageCropper` crops the image with a sliding window.

    :param target_size: desired output size. If size is a sequence like (h, w),
        the output size will be matched to this.
        If size is an int, the output will have the
        same height and width as the `target_size`.
    :param strides: the strides between two neighboring sliding windows.
        `strides` is a sequence like (h, w),
        in which denote the strides on the vertical
        and the horizontal axis.
    :param padding: If False, only patches which are
        fully contained in the input image are included.
        If True, all patches whose starting point
        is inside the input are included,
        and areas outside the input default to zero.
        The `padding` argument has no effect on the size of each patch,
        it determines how many patches are extracted.
        Default is False.
    :param args:  Additional positional arguments
    :param kwargs: Additional keyword arguments
    """

    def __init__(self,
                 target_size: Union[Tuple[int], int] = 16,
                 strides: Tuple[int, int] = (2, 2),
                 padding: bool = False,
                 channel_axis: int = -1,
                 *args,
                 **kwargs):
        """Set constructor"""
        super().__init__(*args, **kwargs)
        self.target_size = target_size
        if len(strides) != 2:
            raise ValueError(f'strides should be a tuple of two integers: {strides}')
        self.stride_h, self.stride_w = strides
        self.padding = padding
        self.channel_axis = channel_axis

    def _add_zero_padding(self, img: 'np.ndarray') -> 'np.ndarray':
        h, w, c = img.shape
        ext_h = self.target_size - h % self.stride_h
        ext_w = self.target_size - w % self.stride_w
        return np.pad(img,
                      ((0, ext_h), (0, ext_w), (0, 0)),
                      mode='constant',
                      constant_values=0)

    @single
    def segment(self, blob: 'np.ndarray', *args, **kwargs) -> List[Dict]:
        """
        Crop the input image array with a sliding window.

        :param blob: the ndarray of the image with the color channel at the last axis
        :return: a list of chunk dicts with the cropped images.
        :param args:  Additional positional arguments
        :param kwargs: Additional keyword arguments
        """
        raw_img = np.copy(blob)
        raw_img = _move_channel_axis(raw_img, self.channel_axis)
        if self.padding:
            raw_img = self._add_zero_padding(blob)
        h, w, c = raw_img.shape
        row_step = raw_img.strides[0]
        col_step = raw_img.strides[1]

        expanded_img = np.lib.stride_tricks.as_strided(
            raw_img,
            shape=(
                1 + int((h - self.target_size) / self.stride_h),
                1 + int((w - self.target_size) / self.stride_w),
                self.target_size,
                self.target_size,
                c
            ),
            strides=(
                row_step * self.stride_h,
                col_step * self.stride_w,
                row_step,
                col_step,
                1), writeable=False)

        bbox_locations = [
            (h * self.stride_h, w * self.stride_w)
            for h in range(expanded_img.shape[0])
            for w in range(expanded_img.shape[1])]

        expanded_img = expanded_img.reshape((-1, self.target_size, self.target_size, c))

        results = []
        for location, _blob in zip(bbox_locations, expanded_img):
            blob = _move_channel_axis(_blob, -1, self.channel_axis)
            results.append(dict(offset=0, weight=1.0, blob=blob.astype('float32'), location=location))
        return results



from jina.executors.rankers import Chunk2DocRanker


class WeightedRanker(Chunk2DocRanker):
    """
    Ranker for multimodal example.

    Will give the scores to chunk data according to weight.
    """

    match_required_keys = None
    query_required_keys = {'weight'}

    def score(self, match_idx, query_chunk_meta, match_chunk_meta, *args, **kwargs):
        """
        Given a set of queries (that may correspond to the chunks of a root level query) and a set of matches
        corresponding to the same parent id, compute the matching score of the common parent of the set of matches.
        Returns a score corresponding to the score of the parent document of the matches in `match_idx`

        :param match_idx: A [N x 4] numpy ``ndarray``, column-wise:
                - ``match_idx[:, 0]``: ``parent_id`` of the matched docs, integer
                - ``match_idx[:, 1]``: ``id`` of the matched chunks, integer
                - ``match_idx[:, 2]``: ``id`` of the query chunks, integer
                - ``match_idx[:, 3]``: distance/metric/score between the query and matched chunks, float.
                All the matches belong to the same `parent`
        :param query_chunk_meta: The meta information of the query chunks, where the key is query chunks' ``chunk_id``,
            the value is extracted by the ``query_required_keys``.
        :param match_chunk_meta: The meta information of the matched chunks, where the key is matched chunks'
            ``chunk_id``, the value is extracted by the ``match_required_keys``.
        :param args: Extra positional arguments
        :param kwargs: Extra keyword arguments
        :return: Return the score associated to the parent id of the matches
        """

        scores = match_idx[[self.COL_QUERY_CHUNK_ID, self.COL_SCORE]]

        weight_score = 0.0
        for k, v in scores:
            vv = 1 / (1 + v)
            weight_score += query_chunk_meta[k]['weight'] * vv

        return weight_score



__copyright__ = "Copyright (c) 2021 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import numpy as np
from jina.executors.decorators import batching, as_ndarray
from jina.executors.encoders import BaseAudioEncoder


class ChromaPitchEncoder(BaseAudioEncoder):
    """
    Segment  audio signal into short chroma frames.

    :class:`ChromaPitchEncoder` is based on chroma spectrograms
    (chromagrams) which represent melodic/harmonic features.
    :class:`ChromaPitchEncoder` encodes an audio signal from a
    `Batch x Signal Length` ndarray into a
    `Batch x Concatenated Features` ndarray.

    :param input_sample_rate: input sampling rate in Hz
        (22050 by default)
    :param hop_length: the number of samples between
        successive chroma frames (512 by default)
    :param args:  Additional positional arguments
    :param kwargs: Additional keyword arguments
    """

    def __init__(self, input_sample_rate: int = 22050, hop_length: int = 512, *args, **kwargs):
        """Set Constructor."""
        super().__init__(*args, **kwargs)
        self.input_sample_rate = input_sample_rate
        self.hop_length = hop_length

    @batching
    @as_ndarray
    def encode(self, content: np.ndarray, *args, **kwargs) -> np.ndarray:
        """
        Craft audio signal of each Chunk into short chroma frames.

        Extract chromagrams for each frame and concatenates Chunk
        frame chromagrams into a single Chunk embedding.

        :param content: a `Batch x Signal Length` ndarray, where
            `Signal Length` is a number of samples
        :return: a `Batch x Concatenated Features` ndarray, where
            `Concatenated Features` is a 12-dimensional feature
            vector times the number of the chroma frames
        :param args:  Additional positional arguments
        :param kwargs: Additional keyword arguments
        """
        from librosa.feature import chroma_cqt
        embeds = []
        for chunk_data in content:
            chromagrams = chroma_cqt(y=chunk_data, sr=self.input_sample_rate, n_chroma=12, hop_length=self.hop_length)
            embeds.append(chromagrams.flatten())
        return embeds



from . import Request


class DumpRequest(Request):
    """A request telling an Indexer to dump its data"""

    @property
    def path(self):
        """


        .. # noqa: DAR102


        .. # noqa: DAR201
        """
        return self.body.path

    @path.setter
    def path(self, value):
        self.body.path = value

    @property
    def shards(self):
        """


        .. # noqa: DAR102


        .. # noqa: DAR201
        """
        return self.body.shards

    @shards.setter
    def shards(self, value):
        self.body.shards = value
